@@ -606,15 +606,18 @@ static void init_uinput(struct avctp *session)
 		debug("AVRCP: uinput initialized for %s", address);
 }
 
-static void avctp_connect_session(struct avctp *session)
+static gboolean avctp_connect_session(struct avctp *session)
 {
 	GIOChannel *io;
 	gboolean value;
 
-	session->state = AVCTP_STATE_CONNECTED;
 	session->dev = manager_find_device(&session->dst, NULL, FALSE);
-	if (!session->dev)
-		return;
+	if (!session->dev) {
+		error("Connecting audio device not known (SDP not completed)");
+		return FALSE;
+	}
+
+	session->state = AVCTP_STATE_CONNECTED;
 
 	session->dev->control->session = session;
 
@@ -638,6 +641,8 @@ static void avctp_connect_session(struct avctp *session)
 			G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 			(GIOFunc) session_cb, session);
 	g_io_channel_unref(io);
+
+	return TRUE;
 }
 
 static void auth_cb(DBusError *derr, void *user_data)
@@ -661,6 +666,7 @@ static void avctp_server_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	struct l2cap_options l2o;
 	struct avctp *session;
 	GIOCondition flags = G_IO_ERR | G_IO_HUP | G_IO_NVAL;
+	struct audio_device *dev;
 	char address[18];
 
 	if (err < 0) {
@@ -680,6 +686,15 @@ static void avctp_server_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 		goto drop;
 	}
 
+	dev = manager_get_device(src, dst);
+	if (!dev) {
+		error("Unable to get audio device object for %s", address);
+		goto drop;
+	}
+
+	if (!dev->control)
+		dev->control = control_init(dev);
+
 	session->state = AVCTP_STATE_CONNECTING;
 	session->sock = g_io_channel_unix_get_fd(chan);
 
@@ -689,7 +704,6 @@ static void avctp_server_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 		err = errno;
 		error("getsockopt(L2CAP_OPTIONS): %s (%d)", strerror(err),
 				err);
-		avctp_unref(session);
 		goto drop;
 	}
 
@@ -710,12 +724,14 @@ static void avctp_server_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	return;
 
 proceed:
-	avctp_connect_session(session);
+	if (!avctp_connect_session(session))
+		goto drop;
 
 	return;
 
 drop:
-	close(session->sock);
+	g_io_channel_close(chan);
+	avctp_unref(session);
 }
 
 static GIOChannel *avctp_server_socket(const bdaddr_t *src, gboolean master)
@@ -739,12 +755,22 @@ static GIOChannel *avctp_server_socket(const bdaddr_t *src, gboolean master)
 static void avctp_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 			const bdaddr_t *dst, gpointer data)
 {
-	struct avctp *session = data;
+	struct control *control = data;
+	struct avctp *session = control->session;
 	struct l2cap_options l2o;
 	socklen_t len;
 	int sk;
 	char address[18];
 
+	if (!session) {
+		debug("avctp_connect_cb: session removed while connecting");
+		g_io_channel_close(chan);
+		g_io_channel_unref(chan);
+		return;
+	}
+
+	ba2str(&session->dst, address);
+
 	if (err < 0) {
 		avctp_unref(session);
 		error("AVCTP connect(%s): %s (%d)", address, strerror(-err),
@@ -752,7 +778,6 @@ static void avctp_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 		return;
 	}
 
-	ba2str(&session->dst, address);
 	debug("AVCTP: connected to %s", address);
 
 	g_io_channel_set_close_on_unref(chan, FALSE);
@@ -763,6 +788,7 @@ static void avctp_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	len = sizeof(l2o);
 	if (getsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &l2o, &len) < 0) {
 		err = errno;
+		g_io_channel_unref(chan);
 		avctp_unref(session);
 		error("getsockopt(L2CAP_OPTIONS): %s (%d)", strerror(err),
 				err);
@@ -780,6 +806,7 @@ static void avctp_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	session->io = g_io_add_watch(chan,
 				G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 				(GIOFunc) session_cb, session);
+	g_io_channel_unref(chan);
 }
 
 gboolean avrcp_connect(struct audio_device *dev)
@@ -801,7 +828,7 @@ gboolean avrcp_connect(struct audio_device *dev)
 	session->state = AVCTP_STATE_CONNECTING;
 
 	err = bt_l2cap_connect(&dev->src, &dev->dst, AVCTP_PSM, 0,
-				avctp_connect_cb, session);
+				avctp_connect_cb, control);
 	if (err < 0) {
 		avctp_unref(session);
 		error("Connect failed. %s(%d)", strerror(-err), -err);
